# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from PIL import ImageChops

from pilgram import util


def test_add() -> None:
    im1 = util.fill((2, 2), (0, 127, 255))
    im2 = util.fill((2, 2), (0, 127, 255))

    actual = util.add(im1, im2)
    expected = util.fill((2, 2), (0, 254, 255))
    assert actual == expected


def test_add2() -> None:
    im1 = util.fill((2, 2), (0, 127, 255))
    im2 = util.fill((2, 2), (0, 127, 255))

    actual = util.add(im1, im2)
    expected = ImageChops.add(im1, im2)
    assert actual == expected



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from PIL import Image, ImageMath
from PIL.ImageMath import _Operand
from PIL.ImageMath import imagemath_convert as _convert
from PIL.ImageMath import imagemath_float as _float

from pilgram.css.blending.alpha import alpha_blend
from pilgram.util import invert


def _color_dodge_image_math(cb: _Operand, cs_inv: _Operand) -> _Operand:
    """Returns ImageMath operands for color dodge blend mode"""
    cb = _float(cb)
    cs_inv = _float(cs_inv)

    cm = ((cb != 0) * (cs_inv == 0) + (cb / cs_inv)) * 255
    return _convert(cm, "L")


def _color_dodge(im1: Image.Image, im2: Image.Image) -> Image.Image:
    """The color dodge blend mode.

    Arguments:
        im1: A backdrop image (RGB).
        im2: A source image (RGB).

    Returns:
        The output image.
    """

    return Image.merge(
        "RGB",
        [
            ImageMath.lambda_eval(
                lambda args: _color_dodge_image_math(args["cb"], args["cs_inv"]),
                cb=cb,
                cs_inv=cs_inv,
            )
            for cb, cs_inv in zip(im1.split(), invert(im2).split(), strict=False)
        ],
    )


def color_dodge(im1: Image.Image, im2: Image.Image) -> Image.Image:
    """Brightens the backdrop color to reflect the source color.

    The color dodge formula is defined as:

        if(Cb == 0)
            B(Cb, Cs) = 0
        else if(Cs == 1)
            B(Cb, Cs) = 1
        else
            B(Cb, Cs) = min(1, Cb / (1 - Cs))

    See the W3C document:
    https://www.w3.org/TR/compositing-1/#blendingcolordodge

    Arguments:
        im1: A backdrop image (RGB or RGBA).
        im2: A source image (RGB or RGBA).

    Returns:
        The output image.
    """

    return alpha_blend(im1, im2, _color_dodge)



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pilgram import util


def test_linear_gradient_mask_prepared_horizontal() -> None:
    w, h = (4, 4)
    mask = util.linear_gradient_mask((w, h))

    assert list(mask.getdata()) == [222, 161, 94, 33] * h
    assert mask.size == (w, h)
    assert mask.mode == "L"


def test_linear_gradient_mask_prepared_vertical() -> None:
    w, h = (4, 4)
    mask = util.linear_gradient_mask((w, h), is_horizontal=False)

    assert list(mask.getdata()) == [222] * w + [161] * w + [94] * w + [33] * w
    assert mask.size == (w, h)
    assert mask.mode == "L"


def test_linear_gradient_mask_start_end() -> None:
    w, h = (4, 4)
    start = 100 / 255
    end = 200 / 255
    mask = util.linear_gradient_mask((w, h), start=start, end=end)

    assert list(mask.getdata()) == [100, 133, 166, 200] * h
    assert mask.size == (w, h)
    assert mask.mode == "L"


def test_linear_gradient_mask_start_end_vertical() -> None:
    w, h = (4, 4)
    start = 100 / 255
    end = 200 / 255
    mask = util.linear_gradient_mask((w, h), start=start, end=end, is_horizontal=False)

    expected = [100] * w + [133] * w + [166] * w + [200] * w

    assert list(mask.getdata()) == expected
    assert mask.size == (w, h)
    assert mask.mode == "L"


def test_linear_gradient_prepared() -> None:
    w, h = (4, 4)
    white = (255,) * 3
    black = (0,) * 3
    gradient = util.linear_gradient((w, h), white, black)
    expected_data = [(c,) * 3 for c in [222, 161, 94, 33]] * h

    assert list(gradient.getdata()) == expected_data
    assert gradient.size == (w, h)
    assert gradient.mode == "RGB"



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
from PIL import Image

from pilgram.types import RGBColor, Size
from pilgram.util.fill import fill
from pilgram.util.invert import invert


def _prepared_linear_gradient_mask(
    size: Size,
    start: float,
    end: float,
    is_horizontal: bool = True,
) -> Image.Image:
    """Returns prepared linear gradient mask."""
    assert end >= 1

    mask = invert(Image.linear_gradient("L"))
    w, h = mask.size
    box = (0, round(h * start), w, round(h / end))
    resized_mask = mask.resize(size, box=box)

    if is_horizontal:
        return resized_mask.rotate(90)
    else:
        return resized_mask


def linear_gradient_mask(
    size: Size,
    start: float = 0,
    end: float = 1,
    is_horizontal: bool = True,
) -> Image.Image:
    """Creates mask image for linear gradient image.

    Arguments:
        size: A tuple of 2 integers. The size of output image.
        start: An optional number. The starting point start.
            The point is left-side when `is_horizontal` is True, top otherwise.
            Defaults to 0.
        end: An optional number. The ending point.
            The point is right-side when `is_horizontal` is True,
            bottom otherwise. Defaults to 1.
        is_horizontal: A optional boolean. The direction of gradient line.
            Left to right if True, top to bottom else.

    Returns:
        The mask image.

    Raises:
        AssertionError: if `size`, `start` and/or `end` have invalid size.
    """

    assert len(size) == 2

    if end >= 1:
        return _prepared_linear_gradient_mask(size, start, end, is_horizontal)

    w, h = size
    start *= 255
    end *= 255

    if is_horizontal:
        row = np.linspace(start, end, num=w, dtype=np.uint8)
        mask = np.tile(row, (h, 1))
    else:
        row = np.linspace(start, end, num=h, dtype=np.uint8)
        mask = np.tile(row, (w, 1)).T

    return Image.fromarray(mask)


def linear_gradient(
    size: Size,
    start: RGBColor,
    end: RGBColor,
    is_horizontal: bool = True,
) -> Image.Image:
    """Creates linear gradient image.

    Arguments:
        size: A tuple of 2 integers. The size of output image.
        start: A tuple of 3 integers. The starting point color.
            The point is left-side when `is_horizontal` is True, top otherwise.
        end: A tuple of 3 integers. The ending point color.
            The point is right-side when `is_horizontal` is True,
            the bottom otherwise.
        is_horizontal: An optional boolean. The direction of gradient line.
            Left to right if True, top to bottom else.

    Returns:
        The output image.

    Raises:
        AssertionError: if `size`, `start` and/or `end` have invalid size.
    """

    assert len(size) == 2
    assert len(start) == 3
    assert len(end) == 3

    im_start = fill(size, start)
    im_end = fill(size, end)
    mask = linear_gradient_mask(size, is_horizontal=is_horizontal)

    return Image.composite(im_start, im_end, mask)



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pilgram.css.blending.color import color
from pilgram.css.blending.color_burn import color_burn
from pilgram.css.blending.color_dodge import color_dodge
from pilgram.css.blending.darken import darken
from pilgram.css.blending.difference import difference
from pilgram.css.blending.exclusion import exclusion
from pilgram.css.blending.hard_light import hard_light
from pilgram.css.blending.hue import hue
from pilgram.css.blending.lighten import lighten
from pilgram.css.blending.multiply import multiply
from pilgram.css.blending.normal import normal
from pilgram.css.blending.overlay import overlay
from pilgram.css.blending.screen import screen
from pilgram.css.blending.soft_light import soft_light

__all__ = [
    "color",
    "color_burn",
    "color_dodge",
    "darken",
    "difference",
    "exclusion",
    "hard_light",
    "hue",
    "lighten",
    "multiply",
    "normal",
    "overlay",
    "screen",
    "soft_light",
]



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pilgram import util
from pilgram.types import BlendingFunction


def _assert_alpha_backdrop_support(blending: BlendingFunction) -> None:
    cb = util.fill((2, 2), (0, 128, 255, 0))
    cs = util.fill((2, 2), (255, 128, 0))
    cr = blending(cb, cs)
    assert cr == cs


def _assert_alpha_source_support(blending: BlendingFunction) -> None:
    cb = util.fill((2, 2), (0, 128, 255))
    cs = util.fill((2, 2), (255, 128, 0, 0))
    cr = blending(cb, cs)
    assert cr == cb


def assert_alpha_support(blending: BlendingFunction) -> None:
    _assert_alpha_backdrop_support(blending)
    _assert_alpha_source_support(blending)



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import math

import numpy as np
from PIL import Image, ImageChops

from pilgram import util
from pilgram.css.blending.alpha import alpha_blend
from pilgram.types import LUT256


def _d_cb(cb: int) -> float:
    """Returns D(Cb) - Cb"""

    cb_float = float(cb) / 255

    if cb_float <= 0.25:
        d = ((16 * cb_float - 12) * cb_float + 4) * cb_float
    else:
        d = math.sqrt(cb_float)

    return round((d - cb_float) * 255)


LUT_1_2_x_cs: LUT256 = [int(util.clip(255 - 2 * i)) for i in range(256)]
LUT_cb_x_1_cb: LUT256 = [int(round(util.clip(i * (1 - i / 255)))) for i in range(256)]
LUT_2_x_cs_1: LUT256 = [int(util.clip(2 * i - 255)) for i in range(256)]
LUT_d_cb: LUT256 = [int(_d_cb(i)) for i in range(256)]


def _soft_light(im1: Image.Image, im2: Image.Image) -> Image.Image:
    """The soft light blend mode.

    Arguments:
        im1: A backdrop image (RGB).
        im2: A source image (RGB).

    Returns:
        The output image.
    """

    _1_2_x_cs = util.apply_lut(im2, LUT_1_2_x_cs)
    cb_x_1_cb = util.apply_lut(im1, LUT_cb_x_1_cb)
    c1 = util.subtract(im1, ImageChops.multiply(_1_2_x_cs, cb_x_1_cb))

    _2_x_cs_1 = util.apply_lut(im2, LUT_2_x_cs_1)
    d_cb = util.apply_lut(im1, LUT_d_cb)
    c2 = util.add(im1, ImageChops.multiply(_2_x_cs_1, d_cb))

    cm = np.where(np.asarray(im2) <= 128, np.asarray(c1), np.asarray(c2))
    return Image.fromarray(cm)


def soft_light(im1: Image.Image, im2: Image.Image) -> Image.Image:
    """Darkens or lightens the colors, depending on the source color value.

    The soft light formula is defined as:

        if(Cs <= 0.5)
            B(Cb, Cs) = Cb - (1 - 2 x Cs) x Cb x (1 - Cb)
        else
            B(Cb, Cs) = Cb + (2 x Cs - 1) x (D(Cb) - Cb)

    where

        if(Cb <= 0.25)
            D(Cb) = ((16 * Cb - 12) x Cb + 4) x Cb
        else
            D(Cb) = sqrt(Cb)

    See the W3C document:
    https://www.w3.org/TR/compositing-1/#blendingsoftlight

    Arguments:
        im1: A backdrop image (RGB or RGBA).
        im2: A source image (RGB or RGBA).

    Returns:
        The output image.
    """

    return alpha_blend(im1, im2, _soft_light)



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from PIL import Image

from pilgram.types import LUT256


def apply_lut(im: Image.Image, lut: LUT256) -> Image.Image:
    """Apply LUT to an image.

    Arguments:
        im: An image.
        lut: A LUT (LookUp Table). The size must be 256.

    Returns:
        The output image.

    Raises:
        ValueError: if `lut` has invalid size.
    """

    if len(lut) != 256:
        raise ValueError(f"A size of LUT must be 256: {len(lut)}")

    return im.point(list(lut) * len(im.getbands()))



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from PIL import ImageChops

from pilgram.util import fill, invert


def test_invert() -> None:
    w, h = (4, 4)
    actual = invert(fill((w, h), (0, 127, 255)))
    expected = fill((w, h), (255, 128, 0))
    assert actual == expected


def test_invert2() -> None:
    w, h = (4, 4)
    actual = invert(fill((w, h), (0, 127, 255)))
    expected = ImageChops.invert(fill((w, h), (0, 127, 255)))
    assert actual == expected



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from pilgram import css, util


def test_contrast() -> None:
    im = util.fill((4, 4), (174, 56, 3))
    contrasted_im = css.contrast(im)

    assert list(contrasted_im.getdata()) == list(im.getdata())
    assert contrasted_im.size == im.size
    assert contrasted_im.mode == im.mode


def test_contrast_1() -> None:
    im = util.fill((4, 4), (174, 56, 3))
    contrasted_im = css.contrast(im, 1)
    contrasted_im2 = css.contrast(im)

    assert list(contrasted_im.getdata()) == list(contrasted_im2.getdata())
    assert contrasted_im.size == im.size
    assert contrasted_im.mode == im.mode


def test_contrast_greater_than_1() -> None:
    im = util.fill((4, 4), (174, 56, 3))
    contrasted_im = css.contrast(im, 2)
    contrasted_im2 = css.contrast(im, 1)

    assert list(contrasted_im.getdata()) != list(contrasted_im2.getdata())
    assert contrasted_im.size == im.size
    assert contrasted_im.mode == im.mode


def test_contrast_0() -> None:
    im = util.fill((4, 4), (174, 56, 3))
    im2 = util.fill((4, 4), (128,) * 3)
    contrasted_im = css.contrast(im, 0)

    assert list(contrasted_im.getdata()) == list(im2.getdata())
    assert contrasted_im.size == im.size
    assert contrasted_im.mode == im.mode


def test_contrast_less_than_0() -> None:
    with pytest.raises(AssertionError):
        im = util.fill((4, 4), (174, 56, 3))
        css.contrast(im, -1)


def test_contrast_hsv() -> None:
    im = util.fill((4, 4), (174, 56, 3))
    im2 = im.convert("HSV")
    contrasted_im = css.contrast(im)
    contrasted_im2 = css.contrast(im2)
    contrasted_im2_rgb = contrasted_im2.convert("RGB")

    assert list(contrasted_im.getdata()) == list(contrasted_im2_rgb.getdata())
    assert contrasted_im2.size == im2.size
    assert contrasted_im2.mode == im2.mode



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from PIL import Image


def invert(im: Image.Image) -> Image.Image:
    """Inverts an image.

    Arguments:
        im: An image.

    Returns:
        The output image.
    """

    # NOTE: `ImageChops.invert` and `im.point` are slower than numpy
    return Image.fromarray(255 - np.asarray(im))



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from PIL import Image

from pilgram import css, util


def rise(im: Image.Image) -> Image.Image:
    """Applies Rise filter.

    Arguments:
        im: An input image.

    Returns:
        The output image.
    """

    cb = util.or_convert(im, "RGB")

    cs1 = util.fill(cb.size, (236, 205, 169, 0.15))
    cm1 = css.blending.multiply(cb, cs1)

    cs2 = util.fill(cb.size, (50, 30, 7, 0.4))
    cm2 = css.blending.multiply(cb, cs2)

    gradient_mask1 = util.radial_gradient_mask(cb.size, length=0.55)
    cm = Image.composite(cm1, cm2, gradient_mask1)

    cs3 = util.fill(cb.size, (232, 197, 152, 0.8))
    cm3 = css.blending.overlay(cm, cs3)

    gradient_mask2 = util.radial_gradient_mask(cb.size, scale=0.9)
    cm_ = Image.composite(cm3, cm, gradient_mask2)
    cr = Image.blend(cm, cm_, 0.6)  # opacity

    cr = css.brightness(cr, 1.05)
    cr = css.sepia(cr, 0.2)
    cr = css.contrast(cr, 0.9)
    cr = css.saturate(cr, 0.9)

    return cr



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from PIL import Image

from pilgram.css.blending.hard_light import hard_light


def overlay(im1: Image.Image, im2: Image.Image) -> Image.Image:
    """Multiplies or screens the colors, depending on the backdrop color value

    The overlay formula is defined as:

        B(Cb, Cs) = HardLight(Cs, Cb)

    See the W3C document:
    https://www.w3.org/TR/compositing-1/#blendingoverlay

    Arguments:
        im1: A backdrop image (RGB or RGBA).
        im2: A source image (RGB or RGBA).

    Returns:
        The output image.
    """

    return hard_light(im2, im1)



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from PIL import Image
from PIL.ImageMath import _Operand
from PIL.ImageMath import imagemath_float as _float
from PIL.ImageMath import imagemath_max as _max
from PIL.ImageMath import imagemath_min as _min

from pilgram.types import RGBOperands


def _min3(c: tuple[_Operand, _Operand, _Operand]) -> _Operand:
    """Returns minimum value of 3 elements as ImageMath operands."""
    r, g, b = c
    return _min(_min(r, g), b)


def _max3(c: tuple[_Operand, _Operand, _Operand]) -> _Operand:
    """Returns maximum value of 3 elements as ImageMath operands."""
    r, g, b = c
    return _max(_max(r, g), b)


def _clip_color(c: RGBOperands) -> RGBOperands:
    """Returns clipped color as ImageMath operands.

    The formula is defined as:

        ClipColor(C)
            L = Lum(C)
            n = min(Cred, Cgreen, Cblue)
            x = max(Cred, Cgreen, Cblue)
            if(n < 0)
                C = L + (((C - L) * L) / (L - n))

            if(x > 1)
                C = L + (((C - L) * (1 - L)) / (x - L))

            return C

    See: https://www.w3.org/TR/compositing-1/#blendingnonseparable

    Arguments:
        c: A tuple of 3 ImageMath operands. The color.

    Returns:
        A tuple of 3 ImageMath operands. The clipped color.
    """

    r, g, b = c

    L = lum(c)
    n = _min3(c)
    x = _max3(c)

    def fn(c: _Operand) -> _Operand:
        # C = L + ((C - L) * L) / (L - n)
        #   = (L * (L - n)) / (L - n) + ((C - L) * L) / (L - n)
        #   = ((L * (L - n)) + ((C - L) * L)) / (L - n)
        #   = (L^2 - nL + CL - L^2) / (L - n)
        #   = (CL - nL) / (L - n)
        #   = (L * (C - n)) / (L - n)
        return (n < 0) * ((L * (c - n)) / (L - n)) + (n >= 0) * c

    def fx(c: _Operand) -> _Operand:
        return (x > 255) * (L + ((c - L) * (255 - L)) / (x - L)) + (x <= 255) * c

    r = fx(fn(r))
    g = fx(fn(g))
    b = fx(fn(b))

    return (r, g, b)


def lum(c: RGBOperands) -> _Operand:
    """Returns luminosity as ImageMath operands.

    The formula is defined as:

        Lum(C) = 0.3 x Cred + 0.59 x Cgreen + 0.11 x Cblue

    See: https://www.w3.org/TR/compositing-1/#blendingnonseparable

    Arguments:
        c: A tuple of 3 ImageMath operands. The color.

    Returns:
        An ImageMath operand. The luminosity.
    """
    r, g, b = c
    return _float(r) * 0.3 + _float(g) * 0.59 + _float(b) * 0.11


def lum_im(im: Image.Image) -> Image.Image:
    """Returns luminosity as image.

    The formula is defined as:

        Lum(C) = 0.3 x Cred + 0.59 x Cgreen + 0.11 x Cblue

    See: https://www.w3.org/TR/compositing-1/#blendingnonseparable

    Arguments:
        im: An input image.

    Returns:
        The luminosity image.
    """
    return im.convert("L")


def set_lum(c: RGBOperands, l1: _Operand) -> RGBOperands:
    """Set luminosity to the color.

    The formula is defined as:

        SetLum(C, l)
            d = l - Lum(C)
            Cred = Cred + d
            Cgreen = Cgreen + d
            Cblue = Cblue + d
            return ClipColor(C)

    See: https://www.w3.org/TR/compositing-1/#blendingnonseparable

    Arguments:
        c: A tuple of 3 ImageMath operands. The color.
        l1: An ImageMath operand. The luminosity to set.

    Returns:
        A tuple of 3 ImageMath operands.
    """

    r, g, b = c

    d = l1 - lum(c)
    r += d
    g += d
    b += d

    return _clip_color((r, g, b))


def set_lum_im(c: RGBOperands, l1: _Operand, l2: _Operand) -> RGBOperands:
    """Set luminosity to the color from image.

    The formula is defined as:

        SetLum(C, l)
            d = l - Lum(C)
            Cred = Cred + d
            Cgreen = Cgreen + d
            Cblue = Cblue + d
            return ClipColor(C)

    See: https://www.w3.org/TR/compositing-1/#blendingnonseparable

    Arguments:
        c: A tuple of 3 ImageMath operands. The color.
        l1: An ImageMath operand. The image of l.
        l2: An ImageMath operand. The image of Lum(C).

    Returns:
        A tuple of 3 ImageMath operands.
    """

    r, g, b = c
    d = l1 - l2

    return _clip_color((r + d, g + d, b + d))


def sat(c: RGBOperands) -> _Operand:
    """Returns saturation as ImageMath operands.

    The formula is defined as:

        Sat(C) = max(Cred, Cgreen, Cblue) - min(Cred, Cgreen, Cblue)

    See: https://www.w3.org/TR/compositing-1/#blendingnonseparable

    Arguments:
        c: A tuple of 3 operands. The color.

    Returns:
        An ImageMath operand. The saturation.
    """

    return _max3(c) - _min3(c)


def set_sat(c: RGBOperands, s: _Operand) -> RGBOperands:
    """Set saturation to the color.

    The formula is defined as:

        SetSat(C, s)
            if(Cmax > Cmin)
                Cmid = (((Cmid - Cmin) x s) / (Cmax - Cmin))
                Cmax = s
            else
                Cmid = Cmax = 0
            Cmin = 0
            return C;

    See: https://www.w3.org/TR/compositing-1/#blendingnonseparable

    Arguments:
        c: A tuple of 3 ImageMath operands. The color.
        s: An ImageMath operand. The saturation to set.

    Returns:
        A tuple of 3 ImageMath operands.
    """

    r, g, b = c

    cmax = _max3(c)
    cmin = _min3(c)
    cmid = r + g + b - cmax - cmin
    new_cmid = ((cmid - cmin) * s) / (cmax - cmin)

    # NOTE: use cmax if cmax == cmid
    cmid_r = (cmax > cmin) * (cmax > cmid) * (cmid == r) * new_cmid
    cmid_g = (cmax > cmin) * (cmax > cmid) * (cmid == g) * new_cmid
    cmid_b = (cmax > cmin) * (cmax > cmid) * (cmid == b) * new_cmid

    cmax_r = (cmax > cmin) * (cmax == r) * s
    cmax_g = (cmax > cmin) * (cmax == g) * s
    cmax_b = (cmax > cmin) * (cmax == b) * s

    return (
        cmid_r + cmax_r,
        cmid_g + cmax_g,
        cmid_b + cmax_b,
    )



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from PIL import ImageChops

from pilgram import util


def test_apply_lut_identity() -> None:
    im = util.fill((2, 2), (0, 127, 255))
    lut_identity = list(range(256))
    assert util.apply_lut(im, lut_identity) == im


def test_apply_lut_invert() -> None:
    im = util.fill((2, 2), (0, 127, 255))
    lut_invert = [255 - i for i in range(256)]
    assert util.apply_lut(im, lut_invert) == ImageChops.invert(im)


def test_apply_lut_value_error() -> None:
    im = util.fill((2, 2), (0, 127, 255))
    lut = [255 - i for i in range(255)]
    with pytest.raises(ValueError):
        util.apply_lut(im, lut)



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import floor

import pytest
from PIL import Image, ImageMath
from PIL.ImageMath import imagemath_convert as _convert

from pilgram import util
from pilgram.css.blending.nonseparable import (
    _clip_color,
    _max3,
    _min3,
    lum,
    lum_im,
    sat,
    set_lum,
    set_sat,
)


def test_min3() -> None:
    im = util.fill((1, 1), (0, 128, 255))
    r, g, b = im.split()
    im_min = ImageMath.lambda_eval(
        lambda args: _convert(_min3((args["r"], args["g"], args["b"])), "L"),
        r=r,
        g=g,
        b=b,
    )

    assert list(im_min.getdata()) == [0]


def test_max3() -> None:
    im = util.fill((1, 1), (0, 128, 255))
    r, g, b = im.split()
    im_max = ImageMath.lambda_eval(
        lambda args: _convert(_max3((args["r"], args["g"], args["b"])), "L"),
        r=r,
        g=g,
        b=b,
    )

    assert list(im_max.getdata()) == [255]


def test_clip_color() -> None:
    im = util.fill((1, 1), (0, 128, 255))
    r, g, b = im.split()
    bands = ImageMath.lambda_eval(
        lambda args: _clip_color((args["r"] - 64, args["g"], args["b"] + 64)),
        r=r,
        g=g,
        b=b,
    )

    expected = [
        [pytest.approx(25.70517158047366, 1e-6)],
        [pytest.approx(106.8796587856024, 1e-6)],
        [pytest.approx(187.63136220320442, 1e-6)],
    ]
    assert [list(band.im.getdata()) for band in bands] == expected


def test_lum() -> None:
    im = util.fill((1, 1), (0, 128, 255))
    r, g, b = im.split()
    im_f = ImageMath.lambda_eval(
        lambda args: lum((args["r"], args["g"], args["b"])), r=r, g=g, b=b
    )
    im_l = im_f.convert("L")

    assert list(im_f.getdata()) == [pytest.approx(103.57, 1e-6)]
    assert list(im_l.getdata()) == [floor(103.57)]


def test_lum_im() -> None:
    im = util.fill((1, 1), (0, 128, 255))
    im_lum = lum_im(im)

    assert list(im_lum.getdata()) == [round(103.57)]


def test_set_lum() -> None:
    im1 = util.fill((1, 1), (0, 128, 255))
    im2 = util.fill((1, 1), (128, 128, 128))
    r1, g1, b1 = im1.split()
    r2, g2, b2 = im2.split()
    bands = ImageMath.lambda_eval(
        lambda args: set_lum(
            (args["r1"], args["g1"], args["b1"]),
            lum((args["r2"], args["g2"], args["b2"])),
        ),
        r1=r1,
        g1=g1,
        b1=b1,
        r2=r2,
        g2=g2,
        b2=b2,
    )

    expected1 = [
        [pytest.approx(41.13881001122631, 1e-6)],
        [pytest.approx(148.48874067225782, 1e-6)],
        [255],
    ]
    assert [list(band.im.getdata()) for band in bands] == expected1

    im_set_lum = Image.merge("RGB", [_convert(band, "L").im for band in bands])
    expected2 = [(floor(41.13881001122631), floor(148.48874067225782), 255)]
    assert list(im_set_lum.getdata()) == expected2


def test_sat() -> None:
    im = util.fill((1, 1), (80, 128, 200))
    r, g, b = im.split()
    im_sat = ImageMath.lambda_eval(
        lambda args: _convert(sat((args["r"], args["g"], args["b"])), "L"),
        r=r,
        g=g,
        b=b,
    )

    assert list(im_sat.getdata()) == [120]


def test_set_sat_cmax_gt_cmin() -> None:
    im1 = util.fill((1, 1), (0, 128, 255))
    im2 = util.fill((1, 1), (64, 96, 128))  # sat = 64
    r1, g1, b1 = im1.split()
    r2, g2, b2 = im2.split()
    bands = ImageMath.lambda_eval(
        lambda args: set_sat(
            (args["r1"], args["g1"], args["b1"]),
            sat((args["r2"], args["g2"], args["b2"])),
        ),
        r1=r1,
        g1=g1,
        b1=b1,
        r2=r2,
        g2=g2,
        b2=b2,
    )

    expected = [
        [0],
        [pytest.approx(32.12549019607843, abs=1)],
        [64],
    ]
    assert [list(band.im.getdata()) for band in bands] == expected


def test_set_sat_cmax_eq_cmid_gt_cmin() -> None:
    im1 = util.fill((1, 1), (0, 128, 128))
    im2 = util.fill((1, 1), (64, 96, 128))  # sat = 64
    r1, g1, b1 = im1.split()
    r2, g2, b2 = im2.split()
    bands = ImageMath.lambda_eval(
        lambda args: set_sat(
            (args["r1"], args["g1"], args["b1"]),
            sat((args["r2"], args["g2"], args["b2"])),
        ),
        r1=r1,
        g1=g1,
        b1=b1,
        r2=r2,
        g2=g2,
        b2=b2,
    )

    expected = [[0], [64], [64]]
    assert [list(band.im.getdata()) for band in bands] == expected


def test_set_sat_cmax_eq_cmin() -> None:
    im1 = util.fill((1, 1), (128, 128, 128))
    im2 = util.fill((1, 1), (64, 96, 128))  # sat = 64
    r1, g1, b1 = im1.split()
    r2, g2, b2 = im2.split()
    bands = ImageMath.lambda_eval(
        lambda args: set_sat(
            (args["r1"], args["g1"], args["b1"]),
            sat((args["r2"], args["g2"], args["b2"])),
        ),
        r1=r1,
        g1=g1,
        b1=b1,
        r2=r2,
        g2=g2,
        b2=b2,
    )

    expected = [[0], [0], [0]]
    assert [list(band.im.getdata()) for band in bands] == expected



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pytest
from PIL import Image

from pilgram import css, util
from pilgram.css.blending.tests.helpers import assert_alpha_support


def test_color() -> None:
    cb_array = np.array(
        [
            [[0] * 3, [127] * 3],
            [[128] * 3, [255] * 3],
        ],
        dtype=np.uint8,
    )
    cb = Image.fromarray(cb_array)
    cs = util.fill((2, 2), (0, 128, 255))
    color = css.blending.color(cb, cs)

    expected = [
        (0, 0, 0),
        (39, 148, 255),
        (41, 148, 255),
        (255, 255, 255),
    ]
    expected_approx = [pytest.approx(c, abs=1) for c in expected]

    assert list(color.getdata()) == expected_approx  # almost eq


def test_color_alpha_support() -> None:
    assert_alpha_support(css.blending.color)



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pilgram import util


def test_radial_gradient_mask_prepared() -> None:
    w, h = (5, 5)
    mask = util.radial_gradient_mask((w, h))

    expected_data = [
        52,
        93,
        111,
        94,
        53,
        93,
        150,
        182,
        151,
        94,
        111,
        182,
        231,
        183,
        113,
        94,
        151,
        183,
        152,
        95,
        53,
        94,
        113,
        95,
        54,
    ]

    # TODO: test rectangle
    assert list(mask.getdata()) == expected_data
    assert mask.size == (w, h)
    assert mask.mode == "L"


def test_radial_gradient_mask_length() -> None:
    w, h = (5, 5)
    mask = util.radial_gradient_mask((w, h), length=0.5)

    expected_data = [
        0,
        107,
        149,
        107,
        0,
        107,
        255,
        255,
        255,
        107,
        149,
        255,
        255,
        255,
        149,
        107,
        255,
        255,
        255,
        107,
        0,
        107,
        149,
        107,
        0,
    ]

    assert list(mask.getdata()) == expected_data
    assert mask.size == (w, h)
    assert mask.mode == "L"


def test_radial_gradient_mask_prepared_scale() -> None:
    w, h = (5, 5)
    mask = util.radial_gradient_mask((w, h), scale=1.5)

    expected_data = [
        118,
        147,
        158,
        148,
        119,
        147,
        187,
        207,
        188,
        148,
        158,
        207,
        239,
        209,
        160,
        148,
        188,
        209,
        189,
        149,
        119,
        148,
        160,
        149,
        120,
    ]

    assert list(mask.getdata()) == expected_data
    assert mask.size == (w, h)
    assert mask.mode == "L"


def test_radial_gradient_mask_position() -> None:
    w, h = (5, 5)
    mask = util.radial_gradient_mask((w, h), center=(0, 0))

    expected_data = [
        255,
        210,
        165,
        120,
        75,
        210,
        191,
        154,
        112,
        69,
        165,
        154,
        128,
        92,
        53,
        120,
        112,
        92,
        64,
        30,
        75,
        69,
        53,
        30,
        0,
    ]

    assert list(mask.getdata()) == expected_data
    assert mask.size == (w, h)
    assert mask.mode == "L"


def test_radial_gradient_mask_length_eq_scale() -> None:
    w, h = (5, 5)
    mask = util.radial_gradient_mask((w, h), length=0.5, scale=0.5)

    expected_data = [
        0,
        0,
        0,
        0,
        0,
        0,
        255,
        255,
        255,
        0,
        0,
        255,
        255,
        255,
        0,
        0,
        255,
        255,
        255,
        0,
        0,
        0,
        0,
        0,
        0,
    ]

    assert list(mask.getdata()) == expected_data
    assert mask.size == (w, h)
    assert mask.mode == "L"


def test_radial_gradient_mask_length_ge_1() -> None:
    w, h = (4, 4)
    mask = util.radial_gradient_mask((w, h), length=1)

    assert list(mask.getdata()) == [255] * (w * h)
    assert mask.size == (w, h)
    assert mask.mode == "L"


def test_radial_gradient_mask_scale_le_0() -> None:
    w, h = (4, 4)
    mask = util.radial_gradient_mask((w, h), scale=0)

    assert list(mask.getdata()) == [0] * (w * h)
    assert mask.size == (w, h)
    assert mask.mode == "L"


def test_radial_gradient() -> None:
    pass  # TODO


def test_radial_gradient_255_to_0() -> None:
    # this case is the same as radial_gradient_mask
    w, h = (5, 5)
    gradient = util.radial_gradient((w, h), [(255, 255, 255), (0, 0, 0)])

    expected = util.radial_gradient_mask((w, h))
    assert list(gradient.getdata(0)) == list(expected.getdata())
    assert list(gradient.getdata(1)) == list(expected.getdata())
    assert list(gradient.getdata(2)) == list(expected.getdata())
    assert gradient.size == (w, h)
    assert gradient.mode == "RGB"



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pilgram import css, util


def test_hue_rotate() -> None:
    im = util.fill((4, 4), (174, 56, 3))
    hue_rotated_im = css.hue_rotate(im)

    assert list(hue_rotated_im.getdata()) == list(im.getdata())
    assert hue_rotated_im.size == im.size
    assert hue_rotated_im.mode == im.mode


def test_hue_rotate_0() -> None:
    im = util.fill((4, 4), (174, 56, 3))
    hue_rotated_im = css.hue_rotate(im)
    hue_rotated_im2 = css.hue_rotate(im, 0)

    assert list(hue_rotated_im.getdata()) == list(hue_rotated_im2.getdata())
    assert hue_rotated_im.size == im.size
    assert hue_rotated_im.mode == im.mode


def test_hue_rotate_360() -> None:
    im = util.fill((4, 4), (174, 56, 3))
    hue_rotated_im = css.hue_rotate(im)
    hue_rotated_im2 = css.hue_rotate(im, 360)

    assert list(hue_rotated_im.getdata()) == list(hue_rotated_im2.getdata())
    assert hue_rotated_im.size == im.size
    assert hue_rotated_im.mode == im.mode


def test_hue_rotate_greater_than_0() -> None:
    im = util.fill((4, 4), (174, 56, 3))
    hue_rotated_im = css.hue_rotate(im, 42)
    hue_rotated_im2 = css.hue_rotate(im)

    assert list(hue_rotated_im.getdata()) != list(hue_rotated_im2.getdata())
    assert hue_rotated_im.size == im.size
    assert hue_rotated_im.mode == im.mode


def test_hue_rotate_less_than_0() -> None:
    im = util.fill((4, 4), (174, 56, 3))
    hue_rotated_im = css.hue_rotate(im, -42)
    hue_rotated_im2 = css.hue_rotate(im)

    assert list(hue_rotated_im.getdata()) != list(hue_rotated_im2.getdata())
    assert hue_rotated_im.size == im.size
    assert hue_rotated_im.mode == im.mode


def test_hue_rotate_hsv() -> None:
    im = util.fill((4, 4), (174, 56, 3))
    im2 = im.convert("HSV")
    hue_rotated_im = css.hue_rotate(im)
    hue_rotated_im2 = css.hue_rotate(im2)
    hue_rotated_im2_rgb = hue_rotated_im2.convert("RGB")

    hue_rotated_im_data = list(hue_rotated_im.getdata())
    hue_rotated_im2_rgb_data = list(hue_rotated_im2_rgb.getdata())

    assert hue_rotated_im_data == hue_rotated_im2_rgb_data
    assert hue_rotated_im2.size == im2.size
    assert hue_rotated_im2.mode == im2.mode



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pilgram.util.add import add
from pilgram.util.apply_lut import apply_lut
from pilgram.util.clip import clip
from pilgram.util.fill import fill
from pilgram.util.invert import invert
from pilgram.util.linear_gradient import linear_gradient, linear_gradient_mask
from pilgram.util.or_convert import or_convert
from pilgram.util.radial_gradient import radial_gradient, radial_gradient_mask
from pilgram.util.subtract import subtract

__all__ = [
    "add",
    "apply_lut",
    "clip",
    "fill",
    "invert",
    "linear_gradient",
    "linear_gradient_mask",
    "or_convert",
    "radial_gradient",
    "radial_gradient_mask",
    "subtract",
]



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from PIL import Image, ImageMath
from PIL.ImageMath import _Operand
from PIL.ImageMath import imagemath_convert as _convert
from PIL.ImageMath import imagemath_float as _float

from pilgram.css.blending.alpha import alpha_blend
from pilgram.css.blending.nonseparable import lum_im, sat, set_lum, set_sat
from pilgram.types import RGBOperands


def _hue_image_math(cb: RGBOperands, cs: RGBOperands, lum_cb: _Operand) -> RGBOperands:
    """Returns ImageMath operands for hue blend mode"""
    cb = (_float(cb[0]), _float(cb[1]), _float(cb[2]))
    cs = (_float(cs[0]), _float(cs[1]), _float(cs[2]))
    lum_cb = _float(lum_cb)

    return set_lum(set_sat(cs, sat(cb)), lum_cb)


def _hue(im1: Image.Image, im2: Image.Image) -> Image.Image:
    """The hue blend mode.

    Arguments:
        im1: A backdrop image (RGB).
        im2: A source image (RGB).

    Returns:
        The output image.
    """

    r1, g1, b1 = im1.split()  # Cb
    r2, g2, b2 = im2.split()  # Cs
    lum_cb = lum_im(im1)  # Lum(Cb)

    bands = ImageMath.lambda_eval(
        lambda args: _hue_image_math(
            (args["r1"], args["g1"], args["b1"]),
            (args["r2"], args["g2"], args["b2"]),
            args["lum_cb"],
        ),
        r1=r1,
        g1=g1,
        b1=b1,
        r2=r2,
        g2=g2,
        b2=b2,
        lum_cb=lum_cb,
    )
    bands = [_convert(band, "L").im for band in bands]

    return Image.merge("RGB", bands)


def hue(im1: Image.Image, im2: Image.Image) -> Image.Image:
    """Creates a color with the hue of the source color
    and the saturation and luminosity of the backdrop color.

    The hue formula is defined as:

        B(Cb, Cs) = SetLum(SetSat(Cs, Sat(Cb)), Lum(Cb))

    See the W3C document:
    https://www.w3.org/TR/compositing-1/#blendinghue

    Arguments:
        im1: A backdrop image (RGB or RGBA).
        im2: A source image (RGB or RGBA).

    Returns:
        The output image.
    """

    return alpha_blend(im1, im2, _hue)



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from PIL import Image, ImageChops

from pilgram.css.blending.alpha import alpha_blend


def _exclusion(im1: Image.Image, im2: Image.Image) -> Image.Image:
    """The exclusion blend mode.

    Arguments:
        im1: A backdrop image (RGB).
        im2: A source image (RGB).

    Returns:
        The output image.
    """

    screen = np.asarray(ImageChops.screen(im1, im2))
    multiply = np.asarray(ImageChops.multiply(im1, im2))

    return Image.fromarray(screen - multiply)


def exclusion(im1: Image.Image, im2: Image.Image) -> Image.Image:
    """Produces an effect like Difference but lower in contrast.

    The exclusion formula is defined as:

        B(Cb, Cs) = Cb + Cs - 2 x Cb x Cs

    See the W3C document:
    https://www.w3.org/TR/compositing-1/#blendingexclusion

    Arguments:
        im1: A backdrop image (RGB or RGBA).
        im2: A source image (RGB or RGBA).

    Returns:
        The output image.
    """

    return alpha_blend(im1, im2, _exclusion)



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pilgram import util


def test_clip_minus_1() -> None:
    assert util.clip(-1) == 0


def test_clip_0() -> None:
    assert util.clip(0) == 0


def test_clip_255() -> None:
    assert util.clip(255) == 255


def test_clip_256() -> None:
    assert util.clip(256) == 255


def test_clip_min_minus_1000() -> None:
    assert util.clip(-1000, a_min=-1000) == -1000
    assert util.clip(-1001, a_min=-1000) == -1000


def test_clip_max_1000() -> None:
    assert util.clip(1000, a_max=1000) == 1000
    assert util.clip(1001, a_max=1000) == 1000



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from PIL import Image

from pilgram import util


def saturate(im: Image.Image, amount: float = 1) -> Image.Image:
    """Saturates image.

    A saturate operation is equivalent to the following matrix operation:

        | R' |     |0.213+0.787s  0.715-0.715s  0.072-0.072s 0  0 |   | R |
        | G' |     |0.213-0.213s  0.715+0.285s  0.072-0.072s 0  0 |   | G |
        | B' |  =  |0.213-0.213s  0.715-0.715s  0.072+0.928s 0  0 | * | B |
        | A' |     |           0            0             0  1  0 |   | A |
        | 1  |     |           0            0             0  0  1 |   | 1 |

    See the W3C document:
    https://www.w3.org/TR/SVG11/filters.html#feColorMatrixValuesAttribute

    Arguments:
        im: An input image.
        amount: An optional number. The filter amount (percentage).
            Defaults to 1.

    Returns:
        The output image.

    Raises:
        AssertionError: if `amount` is less than 0.
    """

    assert amount >= 0

    matrix = (
        0.213 + 0.787 * amount,
        0.715 - 0.715 * amount,
        0.072 - 0.072 * amount,
        0,
        0.213 - 0.213 * amount,
        0.715 + 0.285 * amount,
        0.072 - 0.072 * amount,
        0,
        0.213 - 0.213 * amount,
        0.715 - 0.715 * amount,
        0.072 + 0.928 * amount,
        0,
    )

    saturated = util.or_convert(im, "RGB").convert("RGB", matrix)
    return util.or_convert(saturated, im.mode)



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from collections.abc import Sequence
from functools import reduce
from typing import Any

import numpy as np
from PIL import Image

from pilgram.types import RGBColor, Size
from pilgram.util.fill import fill
from pilgram.util.invert import invert


def _prepared_radial_gradient_mask(size: Size, scale: float = 1) -> Image.Image:
    """Returns prepared radial gradient mask"""

    mask = invert(Image.radial_gradient("L"))

    w, h = mask.size
    xoffset = round((w - w / scale) / 2)
    yoffset = round((h - h / scale) / 2)
    box = (xoffset, yoffset, w - xoffset, h - yoffset)

    return mask.resize(size, box=box)


def radial_gradient_mask(
    size: Size,
    length: float = 0,
    scale: float = 1,
    center: tuple[float, float] = (0.5, 0.5),
) -> Image.Image:
    """Creates mask image for radial gradient image.

    Arguments:
        size: A tuple of 2 integers. The size of mask image.
        length: An optional number. The percentage of inner color stop.
            Defaults to 0.
        scale: An optional number. The percentage of ending shape.
            Defaults to 1.
        center: An optional tuple of two floats.
            The percentage of center position for the circle.
            Defaults to the center (0.5, 0.5).

    Returns:
        The mask image.
    """

    if length >= 1:
        return Image.new("L", size, 255)

    if scale <= 0:
        return Image.new("L", size, 0)

    w, h = size
    cx, cy = center

    # use faster method if possible
    if length == 0 and scale >= 1 and w == h and center == (0.5, 0.5):
        return _prepared_radial_gradient_mask(size, scale)

    rw_left = w * cx
    rw_right = w * (1 - cx)
    rh_top = h * cy
    rh_bottom = h * (1 - cy)

    x = np.linspace(-rw_left, rw_right, w)
    y = np.linspace(-rh_top, rh_bottom, h)[:, None]

    # r is a radius to the farthest-corner
    r = math.sqrt(max(rw_left, rw_right) ** 2 + max(rh_top, rh_bottom) ** 2)
    base = max(scale - length, 0.001)  # avoid a division by zero

    mask = np.sqrt(x**2 + y**2) / r  # distance from center
    mask = (mask - length) / base  # adjust ending shape
    mask = 1 - mask  # invert: distance to center
    mask *= 255
    mask = mask.clip(0, 255)

    return Image.fromarray(np.uint8(mask.round()))


def radial_gradient(
    size: Size,
    colors: Sequence[RGBColor],
    positions: Sequence[float] | None = None,
    **kwargs: Any,
) -> Image.Image:
    """Creates radial gradient image.

    Arguments:
        size: A tuple of 2 integers. The size of output image.
        colors: A sequence of RGB colors.
        positions: An optional sequence of floats.
            The position of color stops.
            If omitted, the positions are equal spacing.

    Returns:
        The output image.

    Raises:
        AssertionError: if `size`, `colors` or `positions` have invalid size.
    """

    assert len(size) == 2
    assert len(colors) >= 2
    for color in colors:
        assert len(color) == 3

    if positions is None:
        positions = list(np.linspace(0, 1, len(colors)))
    else:
        assert len(positions) >= 2
        assert len(colors) == len(positions)

    color_images = [fill(size, color) for color in colors]

    def compose(
        x: tuple[Image.Image, float], y: tuple[Image.Image, float]
    ) -> tuple[Image.Image, float]:
        kwargs_ = kwargs.copy()
        kwargs_["length"] = x[1]
        kwargs_["scale"] = y[1]
        mask = radial_gradient_mask(size, **kwargs_)
        return (Image.composite(x[0], y[0], mask), y[1])

    return reduce(compose, zip(color_images, positions, strict=False))[0]



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from PIL import Image

from pilgram import util


def sepia(im: Image.Image, amount: float = 1) -> Image.Image:
    """Converts image to sepia.

    A sepia operation is equivalent to the following matrix operation:

    | R' |     |0.393+0.607s  0.769-0.769s  0.189-0.189s 0  0 |   | R |
    | G' |     |0.349-0.349s  0.686+0.314s  0.168-0.168s 0  0 |   | G |
    | B' |  =  |0.272-0.272g  0.534-0.534g  0.131+0.869g 0  0 | * | B |
    | A' |     |          0             0             0  1  0 |   | A |
    | 1  |     |          0             0             0  0  1 |   | 1 |

    See the W3C document:
    https://www.w3.org/TR/filter-effects-1/#sepiaEquivalent

    Arguments:
        im: An input image.
        amount: An optional number. The filter amount (percentage).
            Defaults to 1.

    Returns:
        The output image.

    Raises:
        AssertionError: if `amount` is less than 0.
    """

    assert amount >= 0

    amount = 1 - min(amount, 1)
    matrix = (
        0.393 + 0.607 * amount,
        0.769 - 0.769 * amount,
        0.189 - 0.189 * amount,
        0,
        0.349 - 0.349 * amount,
        0.686 + 0.314 * amount,
        0.168 - 0.168 * amount,
        0,
        0.272 - 0.272 * amount,
        0.534 - 0.534 * amount,
        0.131 + 0.869 * amount,
        0,
    )

    sepia_toned = util.or_convert(im, "RGB").convert("RGB", matrix)
    return util.or_convert(sepia_toned, im.mode)



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pilgram._1977 import _1977
from pilgram._version import __version__
from pilgram.aden import aden
from pilgram.brannan import brannan
from pilgram.brooklyn import brooklyn
from pilgram.clarendon import clarendon
from pilgram.earlybird import earlybird
from pilgram.gingham import gingham
from pilgram.hudson import hudson
from pilgram.inkwell import inkwell
from pilgram.kelvin import kelvin
from pilgram.lark import lark
from pilgram.lofi import lofi
from pilgram.maven import maven
from pilgram.mayfair import mayfair
from pilgram.moon import moon
from pilgram.nashville import nashville
from pilgram.perpetua import perpetua
from pilgram.reyes import reyes
from pilgram.rise import rise
from pilgram.slumber import slumber
from pilgram.stinson import stinson
from pilgram.toaster import toaster
from pilgram.valencia import valencia
from pilgram.walden import walden
from pilgram.willow import willow
from pilgram.xpro2 import xpro2

__all__ = [
    "__version__",
    "_1977",
    "aden",
    "brannan",
    "brooklyn",
    "clarendon",
    "earlybird",
    "gingham",
    "hudson",
    "inkwell",
    "kelvin",
    "lark",
    "lofi",
    "maven",
    "mayfair",
    "moon",
    "nashville",
    "perpetua",
    "reyes",
    "rise",
    "slumber",
    "stinson",
    "toaster",
    "valencia",
    "walden",
    "willow",
    "xpro2",
]



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from PIL import Image, ImageChops

from pilgram.types import BlendingFunction
from pilgram.util import add, invert, subtract


def add3(im1: Image.Image, im2: Image.Image, im3: Image.Image) -> Image.Image:
    im1_array = np.asarray(im1, dtype=np.int16)
    im2_array = np.asarray(im2)
    im3_array = np.asarray(im3)
    im_array = im1_array + im2_array + im3_array
    im_array = im_array.clip(0, 255).astype(np.uint8)
    return Image.fromarray(im_array)


def split_alpha(im: Image.Image) -> tuple[Image.Image, Image.Image | None]:
    """Splits alpha channel

    Arguments:
        im: An image (RGB or RGBA).

    Returns:
        A tuple of the RGB and alpha image.
    """

    if im.mode == "RGBA":
        # NOTE: `merge` is slower than `convert` when using Vanilla Pillow
        a = im.split()[3]
        im = im.convert("RGB")
        return im, a
    elif im.mode == "RGB":
        return im, None
    else:
        raise ValueError("Unsupported mode: " + im.mode)


def alpha_to_rgb(im: Image.Image) -> Image.Image:
    """Converts alpha image to rgb image.

    Arguments:
        im: An image (L only).

    Returns:
        A tuple of the RGB and alpha image.
    """
    if im.mode == "L":
        # NOTE: `merge` is slower than `convert` when using Vanilla Pillow
        im = im.convert("RGB")
        return im
    else:
        raise ValueError("Unsupported mode: " + im.mode)


def alpha_blend(
    im1: Image.Image,
    im2: Image.Image,
    blending: BlendingFunction,
) -> Image.Image:
    """Simple alpha blending

    The formula is defined as:

        simple alpha compositing:
        co = cs + cb x (1 - αs)

        written as non-premultiplied:
        αo x Co = αs x Cs + (1 - αs) x αb x Cb

        now substitute the result of blending for Cs:
        αo x Co = αs x ((1 - αb) x Cs + αb x B(Cb, Cs)) + (1 - αs) x αb x Cb
                = αs x (1 - αb) x Cs + αs x αb x B(Cb, Cs) + (1 - αs) x αb x Cb

    See the W3C document:
    https://www.w3.org/TR/compositing-1/#blending

    Arguments:
        im1: A backdrop image (RGB or RGBA).
        im2: A source image (RGB or RGBA).
        blending: The blending method.

    Returns:
        The output image.
    """

    im1, a1 = split_alpha(im1)
    im2, a2 = split_alpha(im2)
    im_blended = blending(im1, im2)

    if a1 is not None and a2 is not None:
        im_blended_alpha = ImageChops.multiply(a1, a2)
        im1_alpha = subtract(a1, im_blended_alpha)
        im2_alpha = subtract(a2, im_blended_alpha)
        c1 = ImageChops.multiply(alpha_to_rgb(im2_alpha), im2)
        c2 = ImageChops.multiply(alpha_to_rgb(im_blended_alpha), im_blended)
        c3 = ImageChops.multiply(alpha_to_rgb(im1_alpha), im1)
        im_blended = add3(c1, c2, c3)
    elif a1 is not None:
        a1_rgb = alpha_to_rgb(a1)
        a1_invert_rgb = alpha_to_rgb(invert(a1))
        im_blended = add(
            ImageChops.multiply(a1_rgb, im_blended),
            ImageChops.multiply(a1_invert_rgb, im2),
        )
    elif a2 is not None:
        a2_rgb = alpha_to_rgb(a2)
        a2_invert_rgb = alpha_to_rgb(invert(a2))
        im_blended = add(
            ImageChops.multiply(a2_rgb, im_blended),
            ImageChops.multiply(a2_invert_rgb, im1),
        )

    return im_blended



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from PIL import Image
from pytest_mock import MockerFixture

from pilgram import util
from pilgram.css.blending.alpha import alpha_blend, alpha_to_rgb, split_alpha


def _normal(cb: Image.Image, cs: Image.Image) -> Image.Image:
    """The normal blend mode"""
    return cs


def test_split_alpha_rgba() -> None:
    im = util.fill((2, 2), (0, 128, 255, 0.5))
    rgb, a = split_alpha(im)
    assert rgb == util.fill((2, 2), (0, 128, 255))
    assert a == Image.new("L", (2, 2), 128)


def test_split_alpha_rgb() -> None:
    im = util.fill((2, 2), (0, 128, 255))
    rgb, a = split_alpha(im)
    assert id(rgb) == id(im)
    assert a is None


def test_split_alpha_unsupported_mode() -> None:
    im = Image.new("L", (2, 2), 128)
    with pytest.raises(ValueError):
        split_alpha(im)


def test_alpha_to_rgb() -> None:
    im = Image.new("L", (2, 2), 128)
    im_rgb = alpha_to_rgb(im)
    assert im_rgb.mode == "RGB"

    r, g, b = im_rgb.split()
    assert r == im
    assert g == im
    assert b == im


def test_alpha_to_rgb_unsupported_mode() -> None:
    im = util.fill((2, 2), (0, 128, 255))
    with pytest.raises(ValueError):
        alpha_to_rgb(im)


def test_alpha_blend_call_blending(mocker: MockerFixture) -> None:
    cb = util.fill((2, 2), (0, 128, 255, 1))
    cs = util.fill((2, 2), (255, 128, 0, 1))

    return_value, _ = split_alpha(cs)
    normal_stub = mocker.Mock(return_value=return_value)
    alpha_blend(cb, cs, normal_stub)

    # TODO: assert with parameters
    normal_stub.assert_called_once()


def test_alpha_blend_normal_rgb_with_rgb() -> None:
    cb = util.fill((2, 2), (0, 128, 255))
    cs = util.fill((2, 2), (255, 128, 0))
    cr = alpha_blend(cb, cs, _normal)
    assert cr == _normal(cb, cs)


def test_alpha_blend_normal_rgb_with_transparent_source() -> None:
    cb = util.fill((2, 2), (0, 128, 255))
    cs = util.fill((2, 2), (255, 128, 0, 0))
    cr = alpha_blend(cb, cs, _normal)
    assert cr == cb


def test_alpha_blend_normal_rgb_with_opaque_source() -> None:
    cb = util.fill((2, 2), (0, 128, 255))
    cs = util.fill((2, 2), (255, 128, 0, 1))
    cr = alpha_blend(cb, cs, _normal)

    expected = split_alpha(cs)[0]
    assert cr == expected


def test_alpha_blend_normal_transparent_backdrop_with_rgb() -> None:
    cb = util.fill((2, 2), (0, 128, 255, 0))
    cs = util.fill((2, 2), (255, 128, 0))
    cr = alpha_blend(cb, cs, _normal)
    assert cr == cs


def test_alpha_blend_normal_opaque_backdrop_with_rgb() -> None:
    cb = util.fill((2, 2), (0, 128, 255, 1))
    cs = util.fill((2, 2), (255, 128, 0))
    cr = alpha_blend(cb, cs, _normal)
    assert cr == cs


def test_alpha_blend_normal_transparent_backdrop_with_transparent_source() -> None:
    cb = util.fill((2, 2), (0, 128, 255, 0))
    cs = util.fill((2, 2), (255, 128, 0, 0))
    cr = alpha_blend(cb, cs, _normal)

    expected = util.fill((2, 2), (0, 0, 0))
    assert cr == expected


def test_alpha_blend_normal_transparent_backdrop_with_opaque_source() -> None:
    cb = util.fill((2, 2), (0, 128, 255, 0))
    cs = util.fill((2, 2), (255, 128, 0, 1))
    cr = alpha_blend(cb, cs, _normal)

    expected = split_alpha(cs)[0]
    assert cr == expected


def test_alpha_blend_normal_opaque_backdrop_with_transparent_source() -> None:
    cb = util.fill((2, 2), (0, 128, 255, 1))
    cs = util.fill((2, 2), (255, 128, 0, 0))
    cr = alpha_blend(cb, cs, _normal)

    expected = split_alpha(cb)[0]
    assert cr == expected


def test_alpha_blend_normal_opaque_backdrop_with_opaque_source() -> None:
    cb = util.fill((2, 2), (0, 128, 255, 1))
    cs = util.fill((2, 2), (255, 128, 0, 1))
    cr = alpha_blend(cb, cs, _normal)

    expected = split_alpha(cs)[0]
    assert cr == expected



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common type definitions for pilgram."""

from collections.abc import Callable, Sequence

from PIL import Image
from PIL.ImageMath import _Operand

# Color types
RGBColor = tuple[int, int, int]
RGBAColor = tuple[int, int, int, float]

# Image size
Size = tuple[int, int]

# Blending function type
BlendingFunction = Callable[[Image.Image, Image.Image], Image.Image]

# ImageMath operand types
RGBOperands = tuple[_Operand, _Operand, _Operand]  # RGB color as ImageMath operands

# LUT types
LUT256 = Sequence[int]  # Look-up table with exactly 256 elements



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from PIL import Image

from pilgram import css, util
from pilgram.css.blending.alpha import split_alpha
from pilgram.css.blending.tests.helpers import assert_alpha_support


def test_normal() -> None:
    cb = util.fill((2, 2), (255, 128, 0))
    cs = util.fill((2, 2), (0, 128, 255))

    actual = css.blending.normal(cb, cs)
    expected = cs
    assert actual == expected


def test_normal2() -> None:
    cb = util.fill((2, 2), (255, 128, 0, 1.0))
    cs_array = np.array(
        [
            [[0, 128, 255, 0.25], [0, 255, 128, 0.5]],
            [[128, 255, 0, 0.75], [128, 0, 255, 0.1]],
        ],
        dtype=np.uint8,
    )
    cs = Image.fromarray(cs_array)

    actual = css.blending.normal(cb, cs)
    expected, _ = split_alpha(Image.alpha_composite(cb, cs))
    assert actual == expected


def test_overlay_alpha_support() -> None:
    assert_alpha_support(css.blending.normal)



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from PIL import Image, ImageMath
from PIL.ImageMath import _Operand
from PIL.ImageMath import imagemath_convert as _convert
from PIL.ImageMath import imagemath_float as _float

from pilgram.css.blending.alpha import alpha_blend
from pilgram.css.blending.nonseparable import lum_im, set_lum_im
from pilgram.types import RGBOperands


def _color_image_math(
    cs: RGBOperands, lum_cb: _Operand, lum_cs: _Operand
) -> RGBOperands:
    """Returns ImageMath operands for color blend mode"""
    cs = (_float(cs[0]), _float(cs[1]), _float(cs[2]))
    lum_cb = _float(lum_cb)
    lum_cs = _float(lum_cs)

    return set_lum_im(cs, lum_cb, lum_cs)


def _color(im1: Image.Image, im2: Image.Image) -> Image.Image:
    """The color blend mode.

    Arguments:
        im1: A backdrop image (RGB).
        im2: A source image (RGB).

    Returns:
        The output image.
    """

    r, g, b = im2.split()  # Cs
    lum_cb = lum_im(im1)  # Lum(Cb)
    lum_cs = lum_im(im2)  # Lum(C) in SetLum

    bands = ImageMath.lambda_eval(
        lambda args: _color_image_math(
            (args["r"], args["g"], args["b"]), args["lum_cb"], args["lum_cs"]
        ),
        r=r,
        g=g,
        b=b,
        lum_cb=lum_cb,
        lum_cs=lum_cs,
    )
    bands = [_convert(band, "L").im for band in bands]

    return Image.merge("RGB", bands)


def color(im1: Image.Image, im2: Image.Image) -> Image.Image:
    """Creates a color with the hue and saturation of the source color
    and the luminosity of the backdrop color.

    The color formula is defined as:

        B(Cb, Cs) = SetLum(Cs, Lum(Cb))

    See the W3C document:
    https://www.w3.org/TR/compositing-1/#blendingcolor

    Arguments:
        im1: A backdrop image (RGB or RGBA).
        im2: A source image (RGB or RGBA).

    Returns:
        The output image.
    """

    return alpha_blend(im1, im2, _color)



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pytest
from PIL import Image

from pilgram import css, util
from pilgram.css.blending.tests.helpers import assert_alpha_support


def test_hue1() -> None:
    cb_array = np.array(
        [
            [[0, 128, 255], [0, 255, 128]],
            [[128, 255, 0], [128, 0, 255]],
        ],
        dtype=np.uint8,
    )
    cb = Image.fromarray(cb_array)
    cs = util.fill((2, 2), (0, 128, 255))
    hue = css.blending.hue(cb, cs)

    expected = [(0, 128, 255), (102, 179, 255), (143, 199, 255), (0, 82, 163)]
    expected_approx = [pytest.approx(c, abs=3) for c in expected]  # TODO

    assert list(hue.getdata()) == expected_approx  # almost eq


def test_hue2() -> None:
    cb = util.fill((2, 2), (0, 128, 255))
    cs_array = np.array(
        [
            [[0, 128, 255], [0, 255, 128]],
            [[128, 255, 0], [128, 0, 255]],
        ],
        dtype=np.uint8,
    )
    cs = Image.fromarray(cs_array)
    hue = css.blending.hue(cb, cs)

    expected = [(0, 128, 255), (0, 160, 80), (70, 139, 0), (153, 50, 255)]
    expected_approx = [pytest.approx(c, abs=1) for c in expected]

    assert list(hue.getdata()) == expected_approx  # almost eq


def test_hue_alpha_support() -> None:
    assert_alpha_support(css.blending.hue)



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from PIL import Image, ImageChops

from pilgram import util
from pilgram.css.blending.alpha import alpha_blend
from pilgram.types import LUT256

LUT_2x: LUT256 = [int(util.clip(2 * i)) for i in range(256)]
LUT_2x_1: LUT256 = [int(util.clip(2 * i - 255)) for i in range(256)]


def _hard_light(im1: Image.Image, im2: Image.Image) -> Image.Image:
    """The hard light blend mode.

    Arguments:
        im1: A backdrop image (RGB).
        im2: A source image (RGB).

    Returns:
        The output image.
    """

    im2_multiply = util.apply_lut(im2, LUT_2x)
    multiply = np.asarray(ImageChops.multiply(im1, im2_multiply))

    im2_screen = util.apply_lut(im2, LUT_2x_1)
    screen = np.asarray(ImageChops.screen(im1, im2_screen))

    cm = np.where(np.asarray(im2) < 128, multiply, screen)
    return Image.fromarray(cm)


def hard_light(im1: Image.Image, im2: Image.Image) -> Image.Image:
    """Multiplies or screens the colors, depending on the source color value

    The hard light formula is defined as:

        if(Cs <= 0.5)
            B(Cb, Cs) = Multiply(Cb, 2 x Cs)
        else
            B(Cb, Cs) = Screen(Cb, 2 x Cs -1)

    See the W3C document:
    https://www.w3.org/TR/compositing-1/#blendinghardlight

    Arguments:
        im1: A backdrop image (RGB or RGBA).
        im2: A source image (RGB or RGBA).

    Returns:
        The output image.
    """

    return alpha_blend(im1, im2, _hard_light)



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pilgram import util


def test_or_convert_same_mode() -> None:
    w, h = (4, 4)
    im = util.fill((w, h), (0, 127, 255))  # RGB
    converted = util.or_convert(im, "RGB")

    assert converted == im  # should be the same instance
    assert converted.mode == "RGB"


def test_or_convert_different_mode() -> None:
    w, h = (4, 4)
    im = util.fill((w, h), (0, 127, 255, 0.5))  # RGBA
    converted = util.or_convert(im, "RGB")

    assert converted != im
    assert converted.mode == "RGB"



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def clip(a: float, a_min: float = 0, a_max: float = 255) -> float:
    """Clips value

    Arguments:
        a: A number. The input value to clip.
        a_min: An optional number. The minimum value. Defaults to 0.
        a_max: An optional number. The maximum value. Defaults to 255.

    Returns:
        The clipped value.
    """

    return min(max(a, a_min), a_max)



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from PIL import Image


def brightness(im: Image.Image, amount: float = 1) -> Image.Image:
    """Adjusts the brightness.

    A brightness operation is equivalent to the following matrix operation:

        | R' |     | c  0  0 |   | R |
        | G' |  =  | 0  c  0 | * | G |
        | B' |     | 0  0  c |   | B |

    See the W3C document:
    https://www.w3.org/TR/filter-effects-1/#brightnessEquivalent

    Arguments:
        im: An input image.
        amount: An optional number. The filter amount (percentage).
            Defaults to 1.

    Returns:
        The output image.

    Raises:
        AssertionError: if `amount` is less than 0.
    """

    assert amount >= 0

    return im.point(lambda x: round(x * amount))



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from PIL import Image

from pilgram import util


def hue_rotate(im: Image.Image, deg: float = 0) -> Image.Image:
    """Applies hue rotation.

    A hue rotate operation is equivalent to the following matrix operation:

        | R' |     | a00  a01  a02  0  0 |   | R |
        | G' |     | a10  a11  a12  0  0 |   | G |
        | B' |  =  | a20  a21  a22  0  0 | * | B |
        | A' |     | 0    0    0    1  0 |   | A |
        | 1  |     | 0    0    0    0  1 |   | 1 |

    where

        | a00 a01 a02 |    [+0.213 +0.715 +0.072]
        | a10 a11 a12 | =  [+0.213 +0.715 +0.072] +
        | a20 a21 a22 |    [+0.213 +0.715 +0.072]
                                [+0.787 -0.715 -0.072]
        cos(hueRotate value) *  [-0.213 +0.285 -0.072] +
                                [-0.213 -0.715 +0.928]
                                [-0.213 -0.715+0.928]
        sin(hueRotate value) *  [+0.143 +0.140-0.283]
                                [-0.787 +0.715+0.072]

    See the W3C document:
    https://www.w3.org/TR/SVG11/filters.html#feColorMatrixValuesAttribute

    Arguments:
        im: An input image.
        deg: An optional number. The hue rotate value (degrees).
            Defaults to 0.

    Returns:
        The output image.
    """

    cos_hue = math.cos(math.radians(deg))
    sin_hue = math.sin(math.radians(deg))

    matrix = (
        0.213 + cos_hue * 0.787 - sin_hue * 0.213,
        0.715 - cos_hue * 0.715 - sin_hue * 0.715,
        0.072 - cos_hue * 0.072 + sin_hue * 0.928,
        0,
        0.213 - cos_hue * 0.213 + sin_hue * 0.143,
        0.715 + cos_hue * 0.285 + sin_hue * 0.140,
        0.072 - cos_hue * 0.072 - sin_hue * 0.283,
        0,
        0.213 - cos_hue * 0.213 - sin_hue * 0.787,
        0.715 - cos_hue * 0.715 + sin_hue * 0.715,
        0.072 + cos_hue * 0.928 + sin_hue * 0.072,
        0,
    )

    rotated = util.or_convert(im, "RGB").convert("RGB", matrix)
    return util.or_convert(rotated, im.mode)



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from PIL import Image


def add(im1: Image.Image, im2: Image.Image) -> Image.Image:
    """Adds two images.

    Arguments:
        im1: An image.
        im2: An image.

    Returns:
        The output image.
    """

    # NOTE: When using Vanilla Pillow, `ImageChops.add` is slower than numpy
    im1_array = np.asarray(im1, dtype=np.int16)  # avoid overflow
    im2_array = np.asarray(im2)
    im_array = im1_array + im2_array
    im_array = im_array.clip(0, 255).astype(np.uint8)

    return Image.fromarray(im_array)



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from PIL import Image

from pilgram import css, util


def mayfair(im: Image.Image) -> Image.Image:
    """Applies Mayfair filter.

    Arguments:
        im: An input image.

    Returns:
        The output image.
    """

    cb = util.or_convert(im, "RGB")
    size = cb.size
    pos = (0.4, 0.4)

    cs1 = util.fill(size, (255, 255, 255, 0.8))
    cm1 = css.blending.overlay(cb, cs1)

    cs2 = util.fill(size, (255, 200, 200, 0.6))
    cm2 = css.blending.overlay(cb, cs2)

    cs3 = util.fill(size, (17, 17, 17))
    cm3 = css.blending.overlay(cb, cs3)

    mask1 = util.radial_gradient_mask(size, scale=0.3, center=pos)
    cs = Image.composite(cm1, cm2, mask1)

    mask2 = util.radial_gradient_mask(size, length=0.3, scale=0.6, center=pos)
    cs = Image.composite(cs, cm3, mask2)
    cr = Image.blend(cb, cs, 0.4)  # opacity

    cr = css.contrast(cr, 1.1)
    cr = css.saturate(cr, 1.1)

    return cr



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from PIL import Image

from pilgram import css, util


def moon(im: Image.Image) -> Image.Image:
    """Applies Moon filter.

    Arguments:
        im: An input image.

    Returns:
        The output image.
    """

    cb = util.or_convert(im, "RGB")

    cs1 = util.fill(cb.size, (160, 160, 160))
    cs = css.blending.soft_light(cb, cs1)

    cs2 = util.fill(cb.size, (56, 56, 56))
    cr = css.blending.lighten(cs, cs2)

    cr = css.grayscale(cr)
    cr = css.contrast(cr, 1.1)
    cr = css.brightness(cr, 1.1)

    return cr



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from PIL import Image

from pilgram.types import RGBAColor, RGBColor, Size


def fill(size: Size, color: RGBColor | RGBAColor) -> Image.Image:
    """Fills new image with the color.

    Arguments:
        size: A tuple of 2 integers. The size of output image.
        color: A tuple of 3 or 4 integers. The fill color.

    Returns:
        The output image.

    Raises:
        AssertionError: if `size` and/or `color` have invalid size.
    """

    assert len(size) == 2
    assert len(color) in [3, 4]

    if len(color) == 4:
        r, g, b, a = color
        alpha_int = int(round(a * 255))
        color = (r, g, b, alpha_int)

    uniqued = list(set(color))
    cmap = {c: Image.new("L", size, c) for c in uniqued}

    if len(color) == 3:
        r, g, b = color
        return Image.merge("RGB", (cmap[r], cmap[g], cmap[b]))
    else:
        r, g, b, a = color
        return Image.merge("RGBA", (cmap[r], cmap[g], cmap[b], cmap[a]))



# Copyright 2019 Akiomi Kamakura
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from PIL import Image
from pytest_benchmark.fixture import BenchmarkFixture

from pilgram import inkwell, util


def test_inkwell() -> None:
    im = util.fill((32, 32), (255,) * 3)
    inkwell(im)


def test_inkwell_benchmark(benchmark: BenchmarkFixture) -> None:
    with Image.open("examples/mtjimba.jpg") as im:
        benchmark(inkwell, im)
